@@ -1,12 +1,12 @@
-# coding: utf-8
-from __future__ import unicode_literals, division, absolute_import, print_function
+import asyncio
+import warnings
 
 from asn1crypto import pem
 from asn1crypto.x509 import Certificate
 
 from .context import ValidationContext, PKIXValidationParams
 from .errors import ValidationError, PathBuildingError, InvalidCertificateError
-from .validate import validate_path, validate_tls_hostname, validate_usage
+from .validate import async_validate_path, validate_tls_hostname, validate_usage
 from .version import __version__, __version_info__
 from ._errors import pretty_message
 from ._types import type_name, str_cls, byte_cls
@@ -17,6 +17,7 @@ __all__ = [
     '__version_info__',
     'CertificateValidator',
     'ValidationContext',
+    'PKIXValidationParams'
 ]
 
 
@@ -98,7 +99,7 @@ class CertificateValidator():
         self._certificate = end_entity_cert
         self._params = pkix_params
 
-    def _validate_path(self):
+    async def _validate_path(self):
         """
         Builds possible certificate paths and validates them until a valid one
         is found, or all fail.
@@ -123,9 +124,11 @@ class CertificateValidator():
             ))
 
         try:
-            paths = self._context.certificate_registry.build_paths(self._certificate)
-        except (PathBuildingError):
-            if self._certificate.self_signed in set(['yes', 'maybe']):
+            paths = await self._context.certificate_registry.async_build_paths(
+                self._certificate
+            )
+        except PathBuildingError:
+            if self._certificate.self_signed in {'yes', 'maybe'}:
                 raise InvalidCertificateError(pretty_message(
                     '''
                     The X.509 certificate provided is self-signed - "%s"
@@ -136,7 +139,7 @@ class CertificateValidator():
 
         for candidate_path in paths:
             try:
-                validate_path(self._context, candidate_path, self._params)
+                await async_validate_path(self._context, candidate_path, self._params)
                 self._path = candidate_path
                 return
             except (ValidationError) as e:
@@ -160,6 +163,76 @@ class CertificateValidator():
         Validates the certificate path and that the certificate is valid for
         the key usage and extended key usage purposes specified.
 
+        .. deprecated:: 0.17.0
+            Use :meth:`async_validate_usage` instead.
+
+        :param key_usage:
+            A set of unicode strings of the required key usage purposes. Valid
+            values include:
+
+             - "digital_signature"
+             - "non_repudiation"
+             - "key_encipherment"
+             - "data_encipherment"
+             - "key_agreement"
+             - "key_cert_sign"
+             - "crl_sign"
+             - "encipher_only"
+             - "decipher_only"
+
+        :param extended_key_usage:
+            A set of unicode strings of the required extended key usage
+            purposes. These must be either dotted number OIDs, or one of the
+            following extended key usage purposes:
+
+             - "server_auth"
+             - "client_auth"
+             - "code_signing"
+             - "email_protection"
+             - "ipsec_end_system"
+             - "ipsec_tunnel"
+             - "ipsec_user"
+             - "time_stamping"
+             - "ocsp_signing"
+             - "wireless_access_points"
+
+            An example of a dotted number OID:
+
+             - "1.3.6.1.5.5.7.3.1"
+
+        :param extended_optional:
+            A bool - if the extended_key_usage extension may be ommited and still
+            considered valid
+
+        :raises:
+            pyhanko_certvalidator.errors.PathValidationError - when an error occurs validating the path
+            pyhanko_certvalidator.errors.RevokedError - when the certificate or another certificate in its path has been revoked
+            pyhanko_certvalidator.errors.InvalidCertificateError - when the certificate is not valid for the usages specified
+
+        :return:
+            A pyhanko_certvalidator.path.ValidationPath object of the validated
+            certificate validation path
+        """
+
+        warnings.warn(
+            "'validate_usage' is deprecated, use "
+            "'async_validate_usage' instead",
+            DeprecationWarning
+        )
+
+        loop = asyncio.get_event_loop()
+        return loop.run_until_complete(
+            self.async_validate_usage(
+                key_usage, extended_key_usage, extended_optional
+            )
+        )
+
+    async def async_validate_usage(self, key_usage, extended_key_usage=None,
+                                   extended_optional=False):
+        """
+        Validates the certificate path and that the certificate is valid for
+        the key usage and extended key usage purposes specified.
+
         :param key_usage:
             A set of unicode strings of the required key usage purposes. Valid
             values include:
@@ -208,7 +281,7 @@ class CertificateValidator():
             certificate validation path
         """
 
-        self._validate_path()
+        await self._validate_path()
         validate_usage(
             self._context,
             self._certificate,
@@ -224,6 +297,36 @@ class CertificateValidator():
         the hostname provided and that the certificate is valid for the purpose
         of a TLS connection.
 
+        .. deprecated:: 0.17.0
+            Use :meth:`async_validate_tls` instead.
+
+        :param hostname:
+            A unicode string of the TLS server hostname
+
+        :raises:
+            pyhanko_certvalidator.errors.PathValidationError - when an error occurs validating the path
+            pyhanko_certvalidator.errors.RevokedError - when the certificate or another certificate in its path has been revoked
+            pyhanko_certvalidator.errors.InvalidCertificateError - when the certificate is not valid for TLS or the hostname
+
+        :return:
+            A pyhanko_certvalidator.path.ValidationPath object of the validated
+            certificate validation path
+        """
+
+        warnings.warn(
+            "'validate_tls' is deprecated, use 'async_validate_tls' instead",
+            DeprecationWarning
+        )
+
+        loop = asyncio.get_event_loop()
+        return loop.run_until_complete(self.async_validate_tls(hostname))
+
+    async def async_validate_tls(self, hostname):
+        """
+        Validates the certificate path, that the certificate is valid for
+        the hostname provided and that the certificate is valid for the purpose
+        of a TLS connection.
+
         :param hostname:
             A unicode string of the TLS server hostname
 
@@ -237,6 +340,6 @@ class CertificateValidator():
             certificate validation path
         """
 
-        self._validate_path()
+        await self._validate_path()
         validate_tls_hostname(self._context, self._certificate, hostname)
         return self._path
